LIBRARY
EXPORTS
vtlInitialize
vtlClose
vtlCalcTongueRootAutomatically
vtlGetVersion
vtlGetConstants
vtlGetTractParamInfo
vtlGetGlottisParamInfo
vtlGetGlottisParams
vtlGetTractParams
vtlExportTractSvg
vtlTractToTube
vtlGetDefaultTransferFunctionOptions
vtlGetTransferFunction
vtlInputTractToLimitedTract
vtlSynthesisReset
vtlSynthesisAddTube
vtlSynthesisAddTract
vtlSynthBlock
vtlApiTest
vtlSegmentSequenceToGesturalScore
vtlGesturalScoreToAudio
vtlGesturalScoreToTractSequence
vtlGetGesturalScoreDuration
vtlTractSequenceToAudio
