// Copyright (C) 2006  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_PIPe_
#define DLIB_PIPe_ 

#include "pipe/pipe_kernel_1.h"


#endif // DLIB_PIPe_

