// Copyright (C) 2003  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_LINKEr_
#define DLIB_LINKEr_

#include "linker/linker_kernel_1.h"

#endif // DLIB_LINKEr_

